#import "AgoraRtcEnginePlugin.h"
#import <AgoraRtcEngineKit/AgoraRtcEngineKit.h>

@interface AgoraRendererView()
@property (nonatomic, strong) UIView *renderView;
@property (nonatomic, assign) int64_t viewId;
@end

@implementation AgoraRendererView
- (instancetype)initWithFrame:(CGRect)frame viewIdentifier:(int64_t)viewId {
  if (self = [super init]) {
    self.renderView = [[UIView alloc] initWithFrame:frame];
    self.renderView.backgroundColor = [UIColor blackColor];
    self.viewId = viewId;
  }
  return self;
}

- (nonnull UIView *)view {
  return self.renderView;
}
@end

@interface AgoraRenderViewFactory : NSObject<FlutterPlatformViewFactory>
@end

@interface AgoraRtcEnginePlugin()<AgoraRtcEngineDelegate>
@property (strong, nonatomic) AgoraRtcEngineKit *agoraRtcEngine;
@property (strong, nonatomic) FlutterMethodChannel *methodChannel;
@property (strong, nonatomic) NSMutableDictionary *rendererViews;
@end

@implementation AgoraRtcEnginePlugin
#pragma mark - renderer views
+ (instancetype)sharedPlugin {
  static AgoraRtcEnginePlugin* plugin = nil;
  static dispatch_once_t onceToken;
  dispatch_once(&onceToken, ^{
    plugin = [[AgoraRtcEnginePlugin alloc] init];
  });
  return plugin;
}

- (NSMutableDictionary *)rendererViews {
  if (!_rendererViews) {
    _rendererViews = [[NSMutableDictionary alloc] init];
  }
  return _rendererViews;
}

+ (void)addView:(UIView *)view id:(NSNumber *)viewId {
  if (!viewId) {
    return;
  }
  if (view) {
    [[[AgoraRtcEnginePlugin sharedPlugin] rendererViews] setObject:view forKey:viewId];
  } else {
    [self removeViewForId:viewId];
  }
}

+ (void)removeViewForId:(NSNumber *)viewId {
  if (!viewId) {
    return;
  }
  [[[AgoraRtcEnginePlugin sharedPlugin] rendererViews] removeObjectForKey:viewId];
}

+ (UIView *)viewForId:(NSNumber *)viewId {
  if (!viewId) {
    return nil;
  }
  return [[[AgoraRtcEnginePlugin sharedPlugin] rendererViews] objectForKey:viewId];
}

#pragma mark - Flutter
+ (void)registerWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
  FlutterMethodChannel* channel = [FlutterMethodChannel
                                   methodChannelWithName:@"agora_rtc_engine"
                                   binaryMessenger:[registrar messenger]];
  AgoraRtcEnginePlugin* instance = [[AgoraRtcEnginePlugin alloc] init];
  instance.methodChannel = channel;
  [registrar addMethodCallDelegate:instance channel:channel];
  
  AgoraRenderViewFactory *fac = [[AgoraRenderViewFactory alloc] init];
  [registrar registerViewFactory:fac withId:@"AgoraRendererView"];
}

- (UIColor *) UIColorFromRGB:(NSUInteger)rgbValue {
  return [UIColor colorWithRed:((float)((rgbValue & 0xFF0000) >> 16))/255.0 green:((float)((rgbValue & 0xFF00) >> 8))/255.0 blue:((float)(rgbValue & 0xFF))/255.0 alpha:1.0];
}

- (AgoraImage *) makeAgoraImage:(NSDictionary *)options {
  AgoraImage *img = [AgoraImage new];
  img.url = [NSURL URLWithString:options[@"url"]];
  
  img.rect = CGRectMake((CGFloat)[options[@"x"] floatValue],
                        (CGFloat)[options[@"y"] floatValue],
                        (CGFloat)[options[@"width"] floatValue],
                        (CGFloat)[options[@"height"] floatValue]);
  return img;
}

- (void)handleMethodCall:(FlutterMethodCall*)call result:(FlutterResult)result {
  NSString *method = call.method;
  NSDictionary *arguments = call.arguments;
  NSLog(@"plugin handleMethodCall: %@, argus: %@", method, arguments);
  
  // Core Methods
  if ([@"create" isEqualToString:method]) {
    NSString *appId = [self stringFromArguments:arguments key:@"appId"];
    self.agoraRtcEngine = [AgoraRtcEngineKit sharedEngineWithAppId:appId delegate:self];
    result(nil);
  } else if ([@"destroy" isEqualToString:method]) {
    self.agoraRtcEngine = nil;
    [AgoraRtcEngineKit destroy];
    result(nil);
  } else if ([@"setChannelProfile" isEqualToString:method]) {
    NSInteger profile = [self intFromArguments:arguments key:@"profile"];
    [self.agoraRtcEngine setChannelProfile:profile];
    result(nil);
  } else if ([@"setClientRole" isEqualToString:method]) {
    NSInteger role = [self intFromArguments:arguments key:@"role"];
    [self.agoraRtcEngine setClientRole:role];
    result(nil);
  } else if([@"switchChannel" isEqualToString:method]) {
    NSString *token = [self stringFromArguments:arguments key:@"token"];
    NSString *channel = [self stringFromArguments:arguments key:@"channelId"];
    NSInteger res = [self.agoraRtcEngine switchChannelByToken:token channelId:channel joinSuccess:nil];
    if (res == 0) {
      result(@{@"result": @(YES)});
    } else {
      result(@{@"result": @(NO), @"error": @(res)});
    }
  } else if ([@"joinChannel" isEqualToString:method]) {
    NSString *token = [self stringFromArguments:arguments key:@"token"];
    NSString *channel = [self stringFromArguments:arguments key:@"channelId"];
    NSString *info = [self stringFromArguments:arguments key:@"info"];
    NSInteger uid = [self intFromArguments:arguments key:@"uid"];
    
    [self.agoraRtcEngine joinChannelByToken:token channelId:channel info:info uid:uid joinSuccess:nil];
    result([NSNumber numberWithBool:YES]);
  } else if ([@"leaveChannel" isEqualToString:method]) {
    BOOL success = (0 == [self.agoraRtcEngine leaveChannel:nil]);
    result([NSNumber numberWithBool:success]);
  } else if ([@"renewToken" isEqualToString:method]) {
    NSString *token = [self stringFromArguments:arguments key:@"token"];
    [self.agoraRtcEngine renewToken:token];
    result(nil);
  } else if ([@"enableWebSdkInteroperability" isEqualToString:method]) {
    BOOL enabled = [self boolFromArguments:arguments key:@"enabled"];
    [self.agoraRtcEngine enableWebSdkInteroperability:enabled];
    result(nil);
  } else if ([@"getConnectionState" isEqualToString:method]) {
    AgoraConnectionStateType state = [self.agoraRtcEngine getConnectionState];
    result([NSNumber numberWithInteger:state]);
  }
  // Core Audio
  else if ([@"enableAudio" isEqualToString:method]) {
    [self.agoraRtcEngine enableAudio];
    result(nil);
  } else if ([@"disableAudio" isEqualToString:method]) {
    [self.agoraRtcEngine disableAudio];
    result(nil);
  } else if ([@"setAudioProfile" isEqualToString:method]) {
    NSInteger profile = [self intFromArguments:arguments key:@"profile"];
    NSInteger scenario = [self intFromArguments:arguments key:@"scenario"];
    [self.agoraRtcEngine setAudioProfile:profile scenario:scenario];
    result(nil);
  } else if ([@"setAudioProfile" isEqualToString:method]) {
    NSInteger profile = [self intFromArguments:arguments key:@"profile"];
    NSInteger scenario = [self intFromArguments:arguments key:@"scenario"];
    [self.agoraRtcEngine setAudioProfile:profile scenario:scenario];
    result(nil);
  } else if ([@"adjustRecordingSignalVolume" isEqualToString:method]) {
    NSInteger volume = [self intFromArguments:arguments key:@"volume"];
    [self.agoraRtcEngine adjustRecordingSignalVolume:volume];
    result(nil);
  } else if ([@"adjustPlaybackSignalVolume" isEqualToString:method]) {
    NSInteger volume = [self intFromArguments:arguments key:@"volume"];
    [self.agoraRtcEngine adjustPlaybackSignalVolume:volume];
    result(nil);
  } else if ([@"enableAudioVolumeIndication" isEqualToString:method]) {
    NSInteger interval = [self intFromArguments:arguments key:@"interval"];
    NSInteger smooth = [self intFromArguments:arguments key:@"smooth"];
    [self.agoraRtcEngine enableAudioVolumeIndication:interval smooth:smooth];
    result(nil);
  } else if ([@"enableLocalAudio" isEqualToString:method]) {
    BOOL enabled = [self boolFromArguments:arguments key:@"enabled"];
    [self.agoraRtcEngine enableLocalAudio:enabled];
    result(nil);
  } else if ([@"muteLocalAudioStream" isEqualToString:method]) {
    BOOL muted = [self boolFromArguments:arguments key:@"muted"];
    [self.agoraRtcEngine muteLocalAudioStream:muted];
    result(nil);
  } else if ([@"muteRemoteAudioStream" isEqualToString:method]) {
    NSInteger uid = [self intFromArguments:arguments key:@"uid"];
    BOOL muted = [self boolFromArguments:arguments key:@"muted"];
    [self.agoraRtcEngine muteRemoteAudioStream:uid mute:muted];
    result(nil);
  } else if ([@"muteAllRemoteAudioStreams" isEqualToString:method]) {
    BOOL muted = [self boolFromArguments:arguments key:@"muted"];
    [self.agoraRtcEngine muteAllRemoteAudioStreams:muted];
    result(nil);
  } else if ([@"setDefaultMuteAllRemoteAudioStreams" isEqualToString:method]) {
    BOOL muted = [self boolFromArguments:arguments key:@"muted"];
    [self.agoraRtcEngine setDefaultMuteAllRemoteAudioStreams:muted];
    result(nil);
  }
  // Core Video
  else if ([@"enableVideo" isEqualToString:method]) {
    [self.agoraRtcEngine enableVideo];
    result(nil);
  } else if ([@"disableVideo" isEqualToString:method]) {
    [self.agoraRtcEngine disableVideo];
    result(nil);
  } else if ([@"setVideoEncoderConfiguration" isEqualToString:method]) {
    NSDictionary *configDic = [self dictionaryFromArguments:arguments key:@"config"];
    AgoraVideoEncoderConfiguration *config = [self videoEncoderConfigurationFromDic:configDic];
    [self.agoraRtcEngine setVideoEncoderConfiguration:config];
    result(nil);
  } else if ([@"removeNativeView" isEqualToString:method]) {
    NSString *viewId = [self stringFromArguments:arguments key:@"viewId"];
    if (viewId.length) {
      [self.rendererViews removeObjectForKey:viewId];
    }
    result(nil);
  } else if ([@"setupLocalVideo" isEqualToString:method]) {
    NSInteger viewId = [self intFromArguments:arguments key:@"viewId"];
    UIView *view = [AgoraRtcEnginePlugin viewForId:@(viewId)];
    NSInteger renderType = [self intFromArguments:arguments key:@"renderMode"];
    
    AgoraRtcVideoCanvas *canvas = [[AgoraRtcVideoCanvas alloc] init];
    canvas.view = view;
    canvas.renderMode = renderType;
    [self.agoraRtcEngine setupLocalVideo:canvas];
    result(nil);
  } else if ([@"setupRemoteVideo" isEqualToString:method]) {
    NSInteger viewId = [self intFromArguments:arguments key:@"viewId"];
    UIView *view = [AgoraRtcEnginePlugin viewForId:@(viewId)];
    NSInteger renderType = [self intFromArguments:arguments key:@"renderMode"];
    NSInteger uid = [self intFromArguments:arguments key:@"uid"];
    
    AgoraRtcVideoCanvas *canvas = [[AgoraRtcVideoCanvas alloc] init];
    canvas.view = view;
    canvas.renderMode = renderType;
    canvas.uid = uid;
    [self.agoraRtcEngine setupRemoteVideo:canvas];
    result(nil);
  } else if ([@"setLocalRenderMode" isEqualToString:method]) {
    NSInteger mode = [self intFromArguments:arguments key:@"mode"];
    [self.agoraRtcEngine setLocalRenderMode:mode];
    result(nil);
  } else if ([@"setLocalVoiceChanger" isEqualToString:method]) {
    NSInteger changer = [self intFromArguments:arguments key:@"changer"];
    [self.agoraRtcEngine setLocalVoiceChanger:changer];
    result(nil);
  } else if ([@"setRemoteRenderMode" isEqualToString:method]) {
    NSInteger uid = [self intFromArguments:arguments key:@"uid"];
    NSInteger mode = [self intFromArguments:arguments key:@"mode"];
    [self.agoraRtcEngine setRemoteRenderMode:uid mode:mode];
    result(nil);
  } else if ([@"startPreview" isEqualToString:method]) {
    [self.agoraRtcEngine startPreview];
    result(nil);
  } else if ([@"stopPreview" isEqualToString:method]) {
    [self.agoraRtcEngine stopPreview];
    result(nil);
  } else if ([@"enableLocalVideo" isEqualToString:method]) {
    BOOL enabled = [self boolFromArguments:arguments key:@"enabled"];
    [self.agoraRtcEngine enableLocalVideo:enabled];
    result(nil);
  } else if ([@"muteLocalVideoStream" isEqualToString:method]) {
    BOOL muted = [self boolFromArguments:arguments key:@"muted"];
    [self.agoraRtcEngine muteLocalVideoStream:muted];
    result(nil);
  } else if ([@"muteRemoteVideoStream" isEqualToString:method]) {
    NSInteger uid = [self intFromArguments:arguments key:@"uid"];
    BOOL muted = [self boolFromArguments:arguments key:@"muted"];
    [self.agoraRtcEngine muteRemoteVideoStream:uid mute:muted];
    result(nil);
  } else if ([@"muteAllRemoteVideoStreams" isEqualToString:method]) {
    BOOL muted = [self boolFromArguments:arguments key:@"muted"];
    [self.agoraRtcEngine muteAllRemoteVideoStreams:muted];
    result(nil);
  } else if ([@"setDefaultMuteAllRemoteVideoStreams" isEqualToString:method]) {
    BOOL muted = [self boolFromArguments:arguments key:@"muted"];
    [self.agoraRtcEngine setDefaultMuteAllRemoteVideoStreams:muted];
    result(nil);
  }
  
  // Video Pre-process and Post-process
  else if ([@"setBeautyEffectOptions" isEqualToString:method]) {
    BOOL enabled = [self boolFromArguments:arguments key:@"enabled"];
    NSDictionary *optionsDic = [self dictionaryFromArguments:arguments key:@"options"];
    AgoraBeautyOptions *options = [self beautyOptionsFromDic:optionsDic];
    [self.agoraRtcEngine setBeautyEffectOptions:enabled options:options];
    result(nil);
  }

  // Audio Routing Controller
  else if ([@"setDefaultAudioRouteToSpeaker" isEqualToString:method]) {
    BOOL defaultToSpeaker = [self boolFromArguments:arguments key:@"defaultToSpeaker"];
    [self.agoraRtcEngine setDefaultAudioRouteToSpeakerphone:defaultToSpeaker];
    result(nil);
  } else if ([@"setEnableSpeakerphone" isEqualToString:method]) {
    BOOL enabled = [self boolFromArguments:arguments key:@"enabled"];
    [self.agoraRtcEngine setEnableSpeakerphone:enabled];
    result(nil);
  } else if ([@"isSpeakerphoneEnabled" isEqualToString:method]) {
    BOOL enable = [self.agoraRtcEngine isSpeakerphoneEnabled];
    result([NSNumber numberWithBool:enable]);
  }
  // Stream Fallback
  else if ([@"setRemoteUserPriority" isEqualToString:method]) {
    NSInteger uid = [self intFromArguments:arguments key:@"uid"];
    AgoraUserPriority priority = [self intFromArguments:arguments key:@"userPriority"];
    [self.agoraRtcEngine setRemoteUserPriority:uid type:priority];
    result(nil);
  } else if ([@"setLocalPublishFallbackOption" isEqualToString:method]) {
    NSInteger option = [self intFromArguments:arguments key:@"option"];
    [self.agoraRtcEngine setLocalPublishFallbackOption:option];
    result(nil);
  } else if ([@"setRemoteSubscribeFallbackOption" isEqualToString:method]) {
    NSInteger option = [self intFromArguments:arguments key:@"option"];
    [self.agoraRtcEngine setRemoteSubscribeFallbackOption:option];
    result(nil);
  }
  // Dual-stream Mode
  else if ([@"enableDualStreamMode" isEqualToString:method]) {
    BOOL enabled = [self boolFromArguments:arguments key:@"enabled"];
    [self.agoraRtcEngine enableDualStreamMode:enabled];
    result(nil);
  } else if ([@"setRemoteVideoStreamType" isEqualToString:method]) {
    NSInteger uid = [self intFromArguments:arguments key:@"uid"];
    NSInteger streamType = [self intFromArguments:arguments key:@"streamType"];
    [self.agoraRtcEngine setRemoteVideoStream:uid type:streamType];
    result(nil);
  } else if ([@"setRemoteDefaultVideoStreamType" isEqualToString:method]) {
    NSInteger streamType = [self intFromArguments:arguments key:@"streamType"];
    [self.agoraRtcEngine setRemoteDefaultVideoStreamType:streamType];
    result(nil);
  }
  // CDN push & pull
  else if ([@"setLiveTranscoding" isEqualToString:method]) {
    NSDictionary *params = [self dictionaryFromArguments:arguments key:@"transcoding"];
    AgoraLiveTranscoding *transcoding = [AgoraLiveTranscoding defaultTranscoding];
    if (params[@"width"] != [NSNull null] && params[@"height"] != [NSNull null]) {
      transcoding.size = CGSizeMake([params[@"width"] doubleValue], [params[@"height"] doubleValue]);
    }
    if (params[@"videoBitrate"] != [NSNull null]) {
      transcoding.videoBitrate = [params[@"videoBitrate"] integerValue];
    }
    if (params[@"videoFramerate"] != [NSNull null]) {
      transcoding.videoFramerate = [params[@"videoFramerate"] integerValue];
    }
    if (params[@"videoGop"] != [NSNull null]) {
      transcoding.videoGop = [params[@"videoGop"] integerValue];
    }
    if (params[@"videoCodecProfile"] != [NSNull null]) {
      transcoding.videoCodecProfile = (AgoraVideoCodecProfileType)[params[@"videoCodecProfile"] integerValue];
    }
    if (params[@"audioCodecProfile"] != [NSNull null]) {
      transcoding.audioCodecProfile = (AgoraAudioCodecProfileType)[params[@"audioCodecProfile"] integerValue];
    }
    if (params[@"audioSampleRate"] != [NSNull null]) {
      transcoding.audioSampleRate = (AgoraAudioSampleRateType)[params[@"audioSampleRate"] integerValue];
    }
    if (params[@"watermark"] != [NSNull null]) {
      transcoding.watermark = [self makeAgoraImage:@{
                                                     @"url": params[@"watermark"][@"url"],
                                                     @"x": params[@"watermark"][@"x"],
                                                     @"y": params[@"watermark"][@"y"],
                                                     @"width": params[@"watermark"][@"width"],
                                                     @"height": params[@"watermark"][@"height"]
                                                     }];
    }
    if (params[@"backgroundImage"] != [NSNull null]) {
      transcoding.backgroundImage = [self makeAgoraImage:@{
                                                           @"url": params[@"backgroundImage"][@"url"],
                                                           @"x": params[@"backgroundImage"][@"x"],
                                                           @"y": params[@"backgroundImage"][@"y"],
                                                           @"width": params[@"backgroundImage"][@"width"],
                                                           @"height": params[@"backgroundImage"][@"height"]
                                                           }];
    }
    
    if (params[@"backgroundColor"] != [NSNull null]) {
      transcoding.backgroundColor = [self UIColorFromRGB:(NSUInteger)[params[@"backgroundColor"] integerValue]];
    }
    
    if (params[@"audioBitrate"] != [NSNull null]) {
      transcoding.audioBitrate = [params[@"audioBitrate"] integerValue];
    }
    
    if (params[@"audioChannels"] != [NSNull null]) {
      transcoding.audioChannels = [params[@"audioChannels"] integerValue];
    }
    
    if (params[@"transcodingUsers"] != [NSNull null]) {
      NSMutableArray<AgoraLiveTranscodingUser*> *transcodingUsers = [NSMutableArray new];
      for (NSDictionary *optionUser in params[@"transcodingUsers"]) {
        AgoraLiveTranscodingUser *liveUser = [AgoraLiveTranscodingUser new];
        liveUser.uid = (NSUInteger)[optionUser[@"uid"] integerValue];
        liveUser.rect = CGRectMake((CGFloat)[optionUser[@"x"] floatValue],
                                   (CGFloat)[optionUser[@"y"] floatValue],
                                   (CGFloat)[optionUser[@"width"] floatValue],
                                   (CGFloat)[optionUser[@"height"] floatValue]);
        liveUser.zOrder = [optionUser[@"zOrder"] integerValue];
        liveUser.alpha = [optionUser[@"alpha"] doubleValue];
        liveUser.audioChannel = [optionUser[@"audioChannel"] integerValue];
        [transcodingUsers addObject:liveUser];
      }
      transcoding.transcodingUsers = transcodingUsers;
    }
    if (params[@"transcodingExtraInfo"] != [NSNull null]) {
      transcoding.transcodingExtraInfo = [params[@"transcodingExtraInfo"] stringValue];
    }
    NSInteger res = [self.agoraRtcEngine setLiveTranscoding:transcoding];
    result(@(res));
  }
  else if ([@"addPublishStreamUrl" isEqualToString:method]) {
    NSString *url = [self stringFromArguments:arguments key:@"url"];
    Boolean enabled = [self boolFromArguments:arguments key:@"enable"];
    NSInteger res = [self.agoraRtcEngine addPublishStreamUrl:url transcodingEnabled:enabled];
    result(@(res));
  }
  else if ([@"removePublishStreamUrl" isEqualToString:method]) {
    NSString *url = [self stringFromArguments:arguments key:@"url"];
    NSInteger res = [self.agoraRtcEngine removePublishStreamUrl:url];
    result(@(res));
  }
  else if ([@"addInjectStreamUrl" isEqualToString:method]) {
    NSString *url = [self stringFromArguments:arguments key:@"url"];
    NSDictionary *config = [self dictionaryFromArguments:arguments key:@"config"];
    AgoraLiveInjectStreamConfig *streamConfig = [AgoraLiveInjectStreamConfig defaultConfig];
    
    if (config[@"width"] != [NSNull null] && config[@"height"] != [NSNull null]) {
      streamConfig.size = CGSizeMake([config[@"width"] doubleValue], [config[@"height"] doubleValue]);
    }
    
    if (config[@"videoGop"] != [NSNull null]) {
      streamConfig.videoGop = [config[@"videoGop"] integerValue];
    }
    
    if (config[@"videoFramerate"] != [NSNull null]) {
      streamConfig.videoFramerate = [config[@"videoFramerate"] integerValue];
    }
    
    if (config[@"videoBitrate"] != [NSNull null]) {
      streamConfig.videoBitrate = [config[@"videoBitrate"] integerValue];
    }
    
    if (config[@"audioBitrate"] != [NSNull null]) {
      streamConfig.audioBitrate = [config[@"audioBitrate"] integerValue];
    }
    
    if (config[@"audioChannels"] != [NSNull null]) {
      streamConfig.audioChannels = [config[@"audioChannels"] integerValue];
    }
    
    if (config[@"audioSampleRate"] != [NSNull null]) {
      streamConfig.audioSampleRate = (AgoraAudioSampleRateType)[config[@"audioSampleRate"] integerValue];
    }
    
    NSInteger res = [self.agoraRtcEngine addInjectStreamUrl:url config:streamConfig];
    result(@(res));
  }
  else if ([@"removeInjectStreamUrl" isEqualToString:method]) {
    NSString *url = [self stringFromArguments:arguments key:@"url"];
    NSInteger res = [self.agoraRtcEngine removeInjectStreamUrl:url];
    result(@(res));
  }
  // Encryption
  else if ([@"setEncryptionSecret" isEqualToString:method]) {
    NSString *secret = [self stringFromArguments:arguments key:@"secret"];
    [self.agoraRtcEngine setEncryptionSecret:secret];
    result(nil);
  } else if ([@"setEncryptionMode" isEqualToString:method]) {
    NSString *encryptionMode = [self stringFromArguments:arguments key:@"encryptionMode"];
    [self.agoraRtcEngine setEncryptionMode:encryptionMode];
    result(nil);
  }
  // Camera Control
  else if ([@"switchCamera" isEqualToString:method]) {
    [self.agoraRtcEngine switchCamera];
    result(nil);
  }
  // Miscellaneous Methods
  else if ([@"getSdkVersion" isEqualToString:method]) {
    NSString *version = [AgoraRtcEngineKit getSdkVersion];
    result(version);
  }
  
  else {
    result(FlutterMethodNotImplemented);
  }
}

- (void)dealloc {
  [self.methodChannel setMethodCallHandler:nil];
}

#pragma mark - delegate
- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didOccurWarning:(AgoraWarningCode)warningCode {
  [self.methodChannel invokeMethod:@"onWarning" arguments:@{@"warn": @(warningCode)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didOccurError:(AgoraErrorCode)errorCode {
  [self.methodChannel invokeMethod:@"onError" arguments:@{@"err": @(errorCode)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didJoinChannel:(NSString * _Nonnull)channel withUid:(NSUInteger)uid elapsed:(NSInteger) elapsed {
  [self.methodChannel invokeMethod:@"onJoinChannelSuccess" arguments:@{@"channel": channel, @"uid": @(uid), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didRejoinChannel:(NSString * _Nonnull)channel withUid:(NSUInteger)uid elapsed:(NSInteger) elapsed {
  [self.methodChannel invokeMethod:@"onRejoinChannelSuccess" arguments:@{@"channel": channel, @"uid": @(uid), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didLeaveChannelWithStats:(AgoraChannelStats * _Nonnull)stats {
  [self.methodChannel invokeMethod:@"onLeaveChannel" arguments:@{@"stats": [self dicFromChannelStats:stats]}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didClientRoleChanged:(AgoraClientRole)oldRole newRole:(AgoraClientRole)newRole {
  [self.methodChannel invokeMethod:@"onClientRoleChanged" arguments:@{@"oldRole": @(oldRole), @"newRole": @(newRole)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didJoinedOfUid:(NSUInteger)uid elapsed:(NSInteger)elapsed {
  [self.methodChannel invokeMethod:@"onUserJoined" arguments:@{@"uid": @(uid), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didOfflineOfUid:(NSUInteger)uid reason:(AgoraUserOfflineReason)reason {
  [self.methodChannel invokeMethod:@"onUserOffline" arguments:@{@"uid": @(uid), @"reason": @(reason)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine connectionChangedToState:(AgoraConnectionStateType)state reason:(AgoraConnectionChangedReason)reason {
  [self.methodChannel invokeMethod:@"onConnectionStateChanged" arguments:@{@"state": @(state), @"reason": @(reason)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine networkTypeChangedToType:(AgoraNetworkType)type {
  [self.methodChannel invokeMethod:@"onNetworkTypeChanged" arguments:@{@"type": @(type)}];
}

- (void)rtcEngineConnectionDidInterrupted:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onConnectionInterrupted" arguments:nil];
}

- (void)rtcEngineConnectionDidLost:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onConnectionLost" arguments:nil];
}

- (void)rtcEngineConnectionDidBanned:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onConnectionBanned" arguments:nil];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didApiCallExecute:(NSInteger)error api:(NSString * _Nonnull)api result:(NSString * _Nonnull)result {
  [self.methodChannel invokeMethod:@"onApiCallExecuted" arguments:@{@"error": @(error), @"api": api, @"result":  result}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine tokenPrivilegeWillExpire:(NSString * _Nonnull)token {
  [self.methodChannel invokeMethod:@"onTokenPrivilegeWillExpire" arguments:@{@"token": token}];
}

- (void)rtcEngineRequestToken:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onRequestToken" arguments:nil];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didMicrophoneEnabled:(BOOL)enabled {
  [self.methodChannel invokeMethod:@"onMicrophoneEnabled" arguments:@{@"enabled": @(enabled)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine reportAudioVolumeIndicationOfSpeakers:(NSArray<AgoraRtcAudioVolumeInfo *> * _Nonnull)speakers totalVolume:(NSInteger)totalVolume {
  [self.methodChannel invokeMethod:@"onAudioVolumeIndication" arguments:@{@"speakers": [self arrayFromSpeakers:speakers], @"totalVolume": @(totalVolume)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine activeSpeaker:(NSUInteger)speakerUid {
  [self.methodChannel invokeMethod:@"onActiveSpeaker" arguments:@{@"uid": @(speakerUid)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine firstLocalAudioFrame:(NSInteger)elapsed {
  [self.methodChannel invokeMethod:@"onFirstLocalAudioFrame" arguments:@{@"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine firstRemoteAudioFrameOfUid:(NSUInteger)uid elapsed:(NSInteger)elapsed {
  [self.methodChannel invokeMethod:@"onFirstRemoteAudioFrame" arguments:@{@"uid": @(uid), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine firstRemoteAudioFrameDecodedOfUid:(NSUInteger)uid elapsed:(NSInteger)elapsed {
  [self.methodChannel invokeMethod:@"onFirstRemoteAudioDecoded" arguments:@{@"uid": @(uid), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine firstLocalVideoFrameWithSize:(CGSize)size elapsed:(NSInteger)elapsed {
  [self.methodChannel invokeMethod:@"onFirstLocalVideoFrame" arguments:@{@"width": @((int)size.width), @"height": @((int)size.height), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine firstRemoteVideoFrameOfUid:(NSUInteger)uid size:(CGSize)size elapsed:(NSInteger)elapsed {
  [self.methodChannel invokeMethod:@"onFirstRemoteVideoFrame" arguments:@{@"uid": @(uid), @"width": @((int)size.width), @"height": @((int)size.height), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didAudioMuted:(BOOL)muted byUid:(NSUInteger)uid {
  [self.methodChannel invokeMethod:@"onUserMuteAudio" arguments:@{@"muted": @(muted), @"uid": @(uid)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine videoSizeChangedOfUid:(NSUInteger)uid size:(CGSize)size rotation:(NSInteger)rotation {
  [self.methodChannel invokeMethod:@"onVideoSizeChanged" arguments:@{@"uid": @(uid), @"width": @((int)size.width), @"height": @((int)size.height), @"rotation": @(rotation)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit *_Nonnull)engine remoteVideoStateChangedOfUid:(NSUInteger)uid state:(AgoraVideoRemoteState)state reason:(AgoraVideoRemoteStateReason)reason elapsed:(NSInteger)elapsed {
  [self.methodChannel invokeMethod:@"onRemoteVideoStateChanged" arguments:@{@"uid": @(uid), @"state": @(state), @"reason": @(reason), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine localVideoStateChange:(AgoraLocalVideoStreamState)state error:(AgoraLocalVideoStreamError)error {
  [self.methodChannel invokeMethod:@"onLocalVideoStateChanged" arguments:@{@"error": @(error), @"localVideoState": @(state)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit *_Nonnull)engine remoteAudioStateChangedOfUid:(NSUInteger)uid state:(AgoraAudioRemoteState)state reason:(AgoraAudioRemoteStateReason)reason elapsed:(NSInteger)elapsed {
  [self.methodChannel invokeMethod:@"onRemoteAudioStateChanged" arguments:@{@"uid": @(uid), @"state": @(state), @"reason": @(reason), @"elapsed": @(elapsed)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit *_Nonnull)engine localAudioStateChange:(AgoraAudioLocalState)state error:(AgoraAudioLocalError)error {
  [self.methodChannel invokeMethod:@"onLocalAudioStateChanged" arguments:@{@"error": @(error), @"state": @(state)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didLocalPublishFallbackToAudioOnly:(BOOL)isFallbackOrRecover {
  [self.methodChannel invokeMethod:@"onLocalPublishFallbackToAudioOnly" arguments:@{@"isFallbackOrRecover": @(isFallbackOrRecover)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didRemoteSubscribeFallbackToAudioOnly:(BOOL)isFallbackOrRecover byUid:(NSUInteger)uid {
  [self.methodChannel invokeMethod:@"onRemoteSubscribeFallbackToAudioOnly" arguments:@{@"uid": @(uid), @"isFallbackOrRecover": @(isFallbackOrRecover)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didAudioRouteChanged:(AgoraAudioOutputRouting)routing {
  [self.methodChannel invokeMethod:@"onAudioRouteChanged" arguments:@{@"routing": @(routing)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine cameraFocusDidChangedToRect:(CGRect)rect {
  [self.methodChannel invokeMethod:@"onCameraFocusAreaChanged" arguments:@{@"rect": [self dicFromRect:rect]}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine cameraExposureDidChangedToRect:(CGRect)rect {
  [self.methodChannel invokeMethod:@"onCameraExposureAreaChanged" arguments:@{@"rect": [self dicFromRect:rect]}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine audioQualityOfUid:(NSUInteger)uid quality:(AgoraNetworkQuality)quality delay:(NSUInteger)delay lost:(NSUInteger)lost {
  [self.methodChannel invokeMethod:@"onAudioQuality" arguments:@{@"uid": @(uid), @"quality": @(quality), @"delay": @(delay), @"lost": @(lost)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine reportRtcStats:(AgoraChannelStats * _Nonnull)stats {
  [self.methodChannel invokeMethod:@"onRtcStats" arguments:@{@"stats": [self dicFromChannelStats:stats]}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine lastmileQuality:(AgoraNetworkQuality)quality {
  [self.methodChannel invokeMethod:@"onLastmileQuality" arguments:@{@"quality": @(quality)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine networkQuality:(NSUInteger)uid txQuality:(AgoraNetworkQuality)txQuality rxQuality:(AgoraNetworkQuality)rxQuality {
  [self.methodChannel invokeMethod:@"onNetworkQuality" arguments:@{@"uid": @(uid), @"txQuality": @(txQuality), @"rxQuality": @(rxQuality)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine localVideoStats:(AgoraRtcLocalVideoStats * _Nonnull)stats {
  [self.methodChannel invokeMethod:@"onLocalVideoStats" arguments:@{@"stats": [self dicFromLocalVideoStats:stats]}];
}

- (void)rtcEngine:(AgoraRtcEngineKit *_Nonnull)engine localAudioStats:(AgoraRtcLocalAudioStats *_Nonnull)stats {
  [self.methodChannel invokeMethod:@"onLocalAudioStats" arguments:@{@"stats": [self dicFromLocalAudioStats:stats]}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine remoteVideoStats:(AgoraRtcRemoteVideoStats * _Nonnull)stats {
  [self.methodChannel invokeMethod:@"onRemoteVideoStats" arguments:@{@"stats": [self dicFromRemoteVideoStats:stats]}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine remoteAudioStats:(AgoraRtcRemoteAudioStats * _Nonnull)stats {
  [self.methodChannel invokeMethod:@"onRemoteAudioStats" arguments:@{@"stats": [self dicFromRemoteAudioStats:stats]}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine videoTransportStatsOfUid:(NSUInteger)uid delay:(NSUInteger)delay lost:(NSUInteger)lost rxKBitRate:(NSUInteger)rxKBitRate {
  [self.methodChannel invokeMethod:@"onRemoteVideoTransportStats" arguments:@{@"uid": @(uid), @"delay": @(delay), @"lost": @(lost), @"rxKBitRate": @(rxKBitRate)}];
}

- (void)rtcEngineRemoteAudioMixingDidStart:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onRemoteAudioMixingStarted" arguments:nil];
}

- (void)rtcEngineRemoteAudioMixingDidFinish:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onRemoteAudioMixingFinished" arguments:nil];
}

- (void)rtcEngineDidAudioEffectFinish:(AgoraRtcEngineKit * _Nonnull)engine soundId:(NSInteger)soundId {
  [self.methodChannel invokeMethod:@"onAudioEffectFinished" arguments:@{@"soundId": @(soundId)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine streamPublishedWithUrl:(NSString * _Nonnull)url errorCode:(AgoraErrorCode)errorCode {
  [self.methodChannel invokeMethod:@"onStreamPublished" arguments:@{@"url": url, @"error": @(errorCode)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine streamUnpublishedWithUrl:(NSString * _Nonnull)url {
  [self.methodChannel invokeMethod:@"onStreamUnpublished" arguments:@{@"url": url}];
}

- (void)rtcEngineTranscodingUpdated:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onTranscodingUpdated" arguments:nil];
}

- (void)rtcEngine:(AgoraRtcEngineKit *_Nonnull)engine rtmpStreamingChangedToState:(NSString *_Nonnull)url state:(AgoraRtmpStreamingState)state errorCode:(AgoraRtmpStreamingErrorCode)errorCode {
  [self.methodChannel invokeMethod:@"onRtmpStreamingStateChanged" arguments:@{@"url": url, @"error": @(errorCode), @"state": @(state)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine streamInjectedStatusOfUrl:(NSString * _Nonnull)url uid:(NSUInteger)uid status:(AgoraInjectStreamStatus)status {
  [self.methodChannel invokeMethod:@"onStreamInjectedStatus" arguments:@{@"url": url, @"uid": @(uid), @"status": @(status)}];
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine receiveStreamMessageFromUid:(NSUInteger)uid streamId:(NSInteger)streamId data:(NSData * _Nonnull)data {
  NSString *message = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
  if (message) {
    [self.methodChannel invokeMethod:@"onStreamMessage" arguments:@{@"uid": @(uid), @"streamId": @(streamId), @"message": message}];
  }
}

- (void)rtcEngine:(AgoraRtcEngineKit * _Nonnull)engine didOccurStreamMessageErrorFromUid:(NSUInteger)uid streamId:(NSInteger)streamId error:(NSInteger)error missed:(NSInteger)missed cached:(NSInteger)cached {
  [self.methodChannel invokeMethod:@"onStreamMessageError" arguments:@{@"uid": @(uid), @"streamId": @(streamId), @"error": @(error), @"missed": @(missed), @"cached": @(cached)}];
}

- (void)rtcEngineMediaEngineDidLoaded:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onMediaEngineLoadSuccess" arguments:nil];
}

- (void)rtcEngineMediaEngineDidStartCall:(AgoraRtcEngineKit * _Nonnull)engine {
  [self.methodChannel invokeMethod:@"onMediaEngineStartCallSuccess" arguments:nil];
}

#pragma mark - helper
- (NSString *)stringFromArguments:(NSDictionary *)arguments key:(NSString *)key {
  if (![arguments isKindOfClass:[NSDictionary class]]) {
    return nil;
  }
  
  NSString *value = [arguments valueForKey:key];
  if (![value isKindOfClass:[NSString class]]) {
    return nil;
  } else {
    return value;
  }
}

- (NSInteger)intFromArguments:(NSDictionary *)arguments key:(NSString *)key {
  if (![arguments isKindOfClass:[NSDictionary class]]) {
    return 0;
  }
  
  NSNumber *value = [arguments valueForKey:key];
  if (![value isKindOfClass:[NSNumber class]]) {
    return 0;
  } else {
    return [value integerValue];
  }
}

- (double)doubleFromArguments:(NSDictionary *)arguments key:(NSString *)key {
  if (![arguments isKindOfClass:[NSDictionary class]]) {
    return 0;
  }
  
  NSNumber *value = [arguments valueForKey:key];
  if (![value isKindOfClass:[NSNumber class]]) {
    return 0;
  } else {
    return [value doubleValue];
  }
}

- (BOOL)boolFromArguments:(NSDictionary *)arguments key:(NSString *)key {
  if (![arguments isKindOfClass:[NSDictionary class]]) {
    return NO;
  }
  
  NSNumber *value = [arguments valueForKey:key];
  if (![value isKindOfClass:[NSNumber class]]) {
    return NO;
  } else {
    return [value boolValue];
  }
}

- (NSDictionary *)dictionaryFromArguments:(NSDictionary *)arguments key:(NSString *)key {
  if (![arguments isKindOfClass:[NSDictionary class]]) {
    return nil;
  }
  
  NSDictionary *value = [arguments valueForKey:key];
  if (![value isKindOfClass:[NSDictionary class]]) {
    return nil;
  } else {
    return value;
  }
}

- (NSDictionary * _Nonnull)dicFromChannelStats:(AgoraChannelStats * _Nonnull)stats {
  return @{@"totalDuration": @(stats.duration),
           @"txBytes": @(stats.txBytes),
           @"rxBytes": @(stats.rxBytes),
           @"txAudioBytes": @(stats.txAudioBytes),
           @"txVideoBytes": @(stats.txVideoBytes),
           @"rxAudioBytes": @(stats.rxAudioBytes),
           @"rxVideoBytes": @(stats.rxVideoBytes),
           @"txKBitrate": @(stats.txKBitrate),
           @"rxKBitrate": @(stats.rxKBitrate),
           @"txAudioKBitrate": @(stats.txAudioKBitrate),
           @"rxAudioKBitrate": @(stats.rxAudioKBitrate),
           @"txVideoKBitrate": @(stats.txVideoKBitrate),
           @"rxVideoKBitrate": @(stats.rxVideoKBitrate),
           @"lastmileDelay": @(stats.lastmileDelay),
           @"txPacketLossRate": @(stats.txPacketLossRate),
           @"rxPacketLossRate": @(stats.rxPacketLossRate),
           @"users": @(stats.userCount),
           @"cpuAppUsage": @(stats.cpuAppUsage),
           @"cpuTotalUsage": @(stats.cpuTotalUsage)
           };
}

- (NSArray * _Nonnull)arrayFromSpeakers:(NSArray<AgoraRtcAudioVolumeInfo *> * _Nonnull)speakers {
  NSMutableArray *array = [[NSMutableArray alloc] init];
  for (AgoraRtcAudioVolumeInfo *speaker in speakers) {
    [array addObject:@{@"uid": @(speaker.uid),
                       @"volume": @(speaker.volume),
                       }];
  }
  
  return [array copy];
}

- (NSDictionary * _Nonnull)dicFromRect:(CGRect)rect {
  return @{@"x": @(rect.origin.x),
           @"y": @(rect.origin.y),
           @"width": @(rect.size.width),
           @"height": @(rect.size.height),
           };
}

- (NSDictionary * _Nonnull)dicFromLocalVideoStats:(AgoraRtcLocalVideoStats * _Nonnull)stats {
  return @{@"sentBitrate": @(stats.sentBitrate),
           @"sentFrameRate": @(stats.sentFrameRate),
           @"encoderOutputFrameRate": @(stats.encoderOutputFrameRate),
           @"rendererOutputFrameRate": @(stats.rendererOutputFrameRate),
           @"sentTargetBitrate": @(stats.sentTargetBitrate),
           @"sentTargetFrameRate": @(stats.sentTargetFrameRate),
           @"qualityAdaptIndication": @(stats.qualityAdaptIndication),
           @"encodedBitrate": @(stats.encodedBitrate),
           @"encodedFrameWidth": @(stats.encodedFrameWidth),
           @"encodedFrameHeight": @(stats.encodedFrameHeight),
           @"encodedFrameCount": @(stats.encodedFrameCount),
           @"codecType": @(stats.codecType)
           };
}

- (NSDictionary * _Nonnull)dicFromLocalAudioStats:(AgoraRtcLocalAudioStats * _Nonnull)stats {
  return @{@"numChannels": @(stats.numChannels),
           @"sentSampleRate": @(stats.sentSampleRate),
           @"sentBitrate": @(stats.sentBitrate),
           };
}

- (NSDictionary * _Nonnull)dicFromRemoteVideoStats:(AgoraRtcRemoteVideoStats * _Nonnull)stats {
  return @{@"uid": @(stats.uid),
           @"width": @(stats.width),
           @"height": @(stats.height),
           @"receivedBitrate": @(stats.receivedBitrate),
           @"decoderOutputFrameRate": @(stats.decoderOutputFrameRate),
           @"rendererOutputFrameRate": @(stats.rendererOutputFrameRate),
           @"packetLossRate": @(stats.packetLossRate),
           @"rxStreamType": @(stats.rxStreamType),
           @"totalFrozenTime": @(stats.totalFrozenTime),
           @"frozenRate": @(stats.frozenRate),
           };
}

- (NSDictionary * _Nonnull)dicFromRemoteAudioStats:(AgoraRtcRemoteAudioStats * _Nonnull)stats {
  return @{@"uid": @(stats.uid),
           @"quality": @(stats.quality),
           @"networkTransportDelay": @(stats.networkTransportDelay),
           @"jitterBufferDelay": @(stats.jitterBufferDelay),
           @"audioLossRate": @(stats.audioLossRate),
           };
}

- (AgoraBeautyOptions *)beautyOptionsFromDic:(NSDictionary *)dictionary {
  AgoraBeautyOptions *options = [[AgoraBeautyOptions alloc] init];
  if (dictionary) {
    options.lighteningContrastLevel = [self intFromArguments:dictionary key:@"lighteningContrastLevel"];
    options.lighteningLevel = [self doubleFromArguments:dictionary key:@"lighteningLevel"];
    options.smoothnessLevel = [self doubleFromArguments:dictionary key:@"smoothnessLevel"];
    options.rednessLevel = [self doubleFromArguments:dictionary key:@"rednessLevel"];
  }
  return options;
}

- (AgoraVideoEncoderConfiguration *)videoEncoderConfigurationFromDic:(NSDictionary *)dictionary {
  AgoraVideoEncoderConfiguration *configuration = [[AgoraVideoEncoderConfiguration alloc] init];
  if (dictionary) {
    NSInteger width = [self doubleFromArguments:dictionary key:@"width"];
    NSInteger height = [self doubleFromArguments:dictionary key:@"height"];
    NSInteger frameRate = [self intFromArguments:dictionary key:@"frameRate"];
    NSInteger minFrameRate = [self intFromArguments:dictionary key:@"minFrameRate"];
    NSInteger bitrate = [self intFromArguments:dictionary key:@"bitrate"];
    NSInteger minBitrate = [self intFromArguments:dictionary key:@"minBitrate"];
    NSInteger orientationMode = [self intFromArguments:dictionary key:@"orientationMode"];
    NSInteger degradationPreference = [self intFromArguments:dictionary key:@"degradationPreference"];
    
    configuration.dimensions = CGSizeMake(width, height);
    configuration.frameRate = frameRate;
    configuration.minFrameRate = minFrameRate;
    configuration.bitrate = bitrate;
    configuration.minBitrate = minBitrate;
    configuration.orientationMode = orientationMode;
    configuration.degradationPreference = degradationPreference;
  }
  return configuration;
}
@end

@implementation AgoraRenderViewFactory
- (nonnull NSObject<FlutterPlatformView> *)createWithFrame:(CGRect)frame viewIdentifier:(int64_t)viewId arguments:(id _Nullable)args {
  AgoraRendererView *rendererView = [[AgoraRendererView alloc] initWithFrame:frame viewIdentifier:viewId];
  [AgoraRtcEnginePlugin addView:rendererView.view id:@(viewId)];
  return rendererView;
}
@end
